package EPrints::Plugin::Sword::Import::OARJ_Broker;

use strict;

use EPrints::Plugin::Sword::Import;
our @ISA = qw/ EPrints::Plugin::Sword::Import /;

our %SUPPORTED_MIME_TYPES = ("application/zip" => 1,);

our %UNPACK_MIME_TYPES = ("application/zip" => "Sword::Unpack::Sub_Zip",);

sub new
{
  my ($class, %params) = @_;
  my $self = $class->SUPER::new(%params);
  $self->{name} = "SWORD Importer - METS/EPDCX+OA-RJ (Broker version)";
  return $self;
} ## end sub new

###        $opts{file} = $file;
###        $opts{mime_type} = $headers->{content_type};
###        $opts{dataset_id} = $target_collection;
###        $opts{owner_id} = $owner->get_id;
###        $opts{depositor_id} = $depositor->get_id if(defined $depositor);
###        $opts{no_op}   = is this a No-op?
###        $opts{verbose} = is this verbosed?
sub input_file
{
  my ($plugin, %opts) = @_;

  my $session = $plugin->{session};

  my $dir  = $opts{dir};
  my $mime = $opts{mime_type};
  my $file = $opts{file};

  my $NO_OP = $opts{no_op};

  # Let's find the XML file to import:
  unless (defined $SUPPORTED_MIME_TYPES{$mime})
  {
    $plugin->add_verbose("[ERROR] unknown MIME TYPE '$mime'.");
    $plugin->set_status_code(415);
    return undef;
  } ## end unless (defined $SUPPORTED_MIME_TYPES...)

  my $unpacker = $UNPACK_MIME_TYPES{$mime};

  my $tmp_dir;

  if (defined $unpacker)
  {
    $tmp_dir = EPrints::TempDir->new("swordXXX", UNLINK => 0);

    if (!defined $tmp_dir)
    {
      print STDERR
          "\n[SWORD-DEPOSIT] [INTERNAL-ERROR] Failed to create the temp directory!";
      $plugin->add_verbose("[ERROR] failed to create the temp directory.");
      $plugin->set_status_code(500);
      return undef;
    } ## end if (!defined $tmp_dir)

    my $files = $plugin->unpack_files($unpacker, $file, $tmp_dir);

    unless (defined $files)
    {
      $plugin->add_verbose("[ERROR] failed to unpack the files");
      return undef;
    }

    my $candidates = $plugin->get_files_to_import($files, "text/xml");

    if (scalar(@$candidates) == 0)
    {
      $plugin->add_verbose("[ERROR] could not find the XML file");
      $plugin->set_status_code(400);
      return undef;
    } ## end if (scalar(@$candidates...))
    elsif (scalar(@$candidates) > 1)
    {

      # We have more than one possible xml file. Take 'mets.xml' if it exists,
      # otherwise just take the one with the 'highest' name, alphabetic-wise
      my @mets = grep /\bmets\.xml$/, @$candidates;
      
      if (scalar(@mets) == 1)
      {
        $$candidates[0] = $mets[0];
        $plugin->add_verbose(
          "[WARNING] there were more than one XML file in this archive, however I'm using the one called 'mets.xml'"
        );
      } elsif (scalar(@mets) > 1)
      {
        @$candidates = sort { length($a) cmp length($b) } @mets;
        $plugin->add_verbose(
          "[WARNING] there were more than one mets.XML file in this archive, however I'm using the one closest to the top of the 'tree'"
        );
      } else
      {
        @mets = sort @$candidates;
        $$candidates[0] = $mets[0];
        $plugin->add_verbose(
          "[WARNING] there were more than one XML file in this archive. I am using the one with the 'highest' name, alphabetic-wise (closer to 'z' than 'A')"
        );
      } ## end else [ if (scalar(@mets) == 1...)]
    } ## end elsif (scalar(@$candidates...))

    $file = $$candidates[0];
    
  } ## end if (defined $unpacker)

  my $dataset_id   = $opts{dataset_id};
  my $owner_id     = $opts{owner_id};
  my $depositor_id = $opts{depositor_id};

  my $fh;
  if (!open($fh, $file))
  {
    $plugin->add_verbose(
                       "[ERROR] couldnt open the file: '$file' because '$!'");
    $plugin->set_status_code(500);
    return;
  } ## end if (!open($fh, $file))

  my $unpack_dir;
  my $fntmp;

  $fntmp = $file;

  if ($fntmp =~ /^(.*)\/(.*)$/)
  {
    $unpack_dir = $1;
    $fntmp      = $2;
  }

  # needs to read the xml from the file:
  my $xml;
  while (my $d = <$fh>)
  {
    $xml .= $d;
  }
  close $fh;

  my $dataset = $session->get_archive()->get_dataset($dataset_id);

  if (!defined $dataset)
  {
    print STDERR
        "\n[SWORD-METS] [INTERNAL-ERROR] Failed to open the dataset '$dataset_id'.";
    $plugin->add_verbose(
                 "[INTERNAL ERROR] failed to open the dataset '$dataset_id'");
    $plugin->set_status_code(500);
    return;
  } ## end if (!defined $dataset)

  my $dom_doc;
  eval { $dom_doc = EPrints::XML::parse_xml_string($xml); };

  if ($@ || !defined $dom_doc)
  {
    $plugin->add_verbose("[ERROR] failed to parse the xml: '$@'");
    $plugin->set_status_code(400);
    return;
  } ## end if ($@ || !defined $dom_doc...)

  if (!defined $dom_doc)
  {
    $plugin->{status_code} = 400;
    $plugin->add_verbose("[ERROR] failed to parse the xml.");
    return;
  } ## end if (!defined $dom_doc)

  my $dom_top = $dom_doc->getDocumentElement;

  if (lc $dom_top->tagName ne 'mets')
  {
    $plugin->set_status_code(400);
    $plugin->add_verbose(
                     "[ERROR] failed to parse the xml: no <mets> tag found.");
    return;
  } ## end if (lc $dom_top->tagName...)

  # METS Headers (ignored)
  my $mets_hdr = ($dom_top->getElementsByTagName("metsHdr"))[0];

  # METS Descriptive Metadata (main section for us)

  # need to loop on dmdSec:
  my @dmd_sections = $dom_top->getElementsByTagName("dmdSec");
  my $dmd_id;

  my $md_wrap;
  my $found_wrapper = 0;
  foreach my $dmd_sec (@dmd_sections)
  {

    # need to extract xmlData from here
    $md_wrap = ($dmd_sec->getElementsByTagName("mdWrap"))[0];

    next if (!defined $md_wrap);

    next
        if (!(   lc $md_wrap->getAttribute("MDTYPE") eq 'other'
              && defined $md_wrap->getAttribute("OTHERMDTYPE")
              && lc $md_wrap->getAttribute("OTHERMDTYPE") eq 'epdcx'
             )
           );

    $found_wrapper = 1;
    $dmd_id        = $dmd_sec->getAttribute("ID");
    last;
  } ## end foreach my $dmd_sec (@dmd_sections...)

  unless ($found_wrapper)
  {
    $plugin->set_status_code(400);
    $plugin->add_verbose(
      "[ERROR] failed to parse the xml: could not find epdcx <mdWrap> section."
    );
    return;
  } ## end unless ($found_wrapper)

  my $xml_data = ($md_wrap->getElementsByTagName("xmlData"))[0];

  if (!defined $xml_data)
  {
    $plugin->set_status_code(400);
    $plugin->add_verbose(
                  "[ERROR] failed to parse the xml: no <xmlData> tag found.");
    return;
  } ## end if (!defined $xml_data...)

  my $epdata = $plugin->parse_epdcx_xml_data($xml_data);
  return unless (defined $epdata);

  # Add some useful (SWORD) info
  if (defined $depositor_id)
  {
    $epdata->{userid}          = $owner_id;
    $epdata->{sword_depositor} = $depositor_id;
  } else
  {
    $epdata->{userid} = $owner_id;
  }

  $epdata->{eprint_status} = $dataset_id;

  # add the files included in the zip file.
  # This is not a straight "every file is a document" import: EPrints has this
  # weird thing where a "document" can have multiple files. The Broker deals
  # with this by having defining the structure in <structMap>, and putting
  # each document into a <div>. If there are multiple files, the main file is
  # the first file in the set

  # The first thing to do is create a list of all the files & their IDs

  # File Section which will contain optional info about files to import:
  my %files;
  my $file_counter = 0;
  foreach my $file_sec ($dom_top->getElementsByTagName("fileSec"))
  {
    my $file_grp = ($file_sec->getElementsByTagName("fileGrp"))[0];

    $file_sec = $file_grp
        if (defined $file_grp)
        ;    # this is because the <fileGrp> tag is optional

    foreach my $file_div ($file_sec->getElementsByTagName("file"))
    {
      my $file_id  = $file_div->getAttribute("ID");
      my $file_loc = ($file_div->getElementsByTagName("FLocat"))[0];
      if (defined $file_loc)
      {      # yeepee we have a file (maybe)

        my $fn = $file_loc->getAttribute("href");

        unless (defined $fn)
        {

          # to accommodate the gdome XML library:
          $fn = $file_loc->getAttribute("xlink:href");
        } ## end unless (defined $fn)

        next unless (defined $fn);

        $files{$file_id} = $fn;    # list files to get locally
      } ## end if (defined $file_loc)
    } ## end foreach my $file_div ($file_sec...)
  } ## end foreach my $file_sec ($dom_top...)

  unless (scalar keys %files)
  {
    $plugin->add_verbose(
              "[WARNING] no <fileSec> tag found: no files will be imported.");
  }

  # we also need to hang onto details about where the importer unpacked stuff.
  # I know it's nothing to do with files in the METS manifest, however this is
  # probably the best place to keep them :)
  $files{unpack_dir} = $unpack_dir;
  $files{fntmp}      = $fntmp;

  if ($NO_OP)
  {

    # need to send 200 Successful (the deposit handler will generate the XML response)
    $plugin->add_verbose(
                      "[OK] Plugin - import successful (but in No-Op mode).");
    $plugin->set_status_code(200);
    return;
  } ## end if ($NO_OP)

  my $eprint = $dataset->create_object($plugin->{session}, $epdata);

  unless (defined $eprint)
  {
    $plugin->set_status_code(500);
    $plugin->add_verbose("[ERROR] failed to create the EPrint object.");
    return;
  } ## end unless (defined $eprint)

  # Having created a base EPrint, and determined what files we have to add to it,
  # we create a "map" of what files go where from the <structMap> element.
  # $dmd_id contains the ID for the div we're looking for
  foreach my $structMap ($dom_top->getElementsByTagName("structMap"))
  {

    # we want the [first] div that has a dmdid that matches $dmd_id
    my $div;
    my @divs = $structMap->getElementsByTagName("div"); #childNodes;
    foreach my $d (@divs)
    {
      next unless $d->hasAttribute('DMDID');
      my $v = $d->getAttribute('DMDID');
      next unless $v eq $dmd_id;
      $div = $d;
    } ## end foreach my $d (@divs)

    if ($div)
    {

      my $doc_data_ref = {};
      my @nodes        = $div->getElementsByTagName("div"); #childNodes;

      my $document =
          process_struct_elements(\@nodes, \%files, $doc_data_ref, $eprint,
                                  $session);

      if (!defined $document)
      {
        $plugin->add_verbose("[WARNING] Failed to create Document object.");
      }
    } ## end if ($div)
  } ## end foreach my $structMap ($dom_top...)

  if ($plugin->keep_deposited_file())
  {
    if (
       $plugin->attach_deposited_file($eprint, $opts{file}, $opts{mime_type}))
    {
      $plugin->add_verbose("[OK] attached deposited file.");
    } else
    {
      $plugin->add_verbose("[WARNING] failed to attach the deposited file.");
    }
  } ## end if ($plugin->keep_deposited_file...)

  $plugin->add_verbose("[OK] EPrint object created.");

  return $eprint;

} ## end sub input_file

sub process_struct_elements
{
  my ($nodes_ref, $files_hashref, $doc_data_ref, $eprint, $session) = @_;

  my $hash_ref = {};

  my $document;

  my $unpack_dir = $files_hashref->{unpack_dir};

  # Go through all the children
  foreach my $element (@{$nodes_ref})
  {
    my @fptrs = $element->getElementsByTagName("fptr"); #childNodes;

    # check to see what element it is
    if (scalar @fptrs )
    {
          my $doc_data = {};
          foreach my $fptr (@fptrs)
          {
            my $fileid = $fptr->getAttribute("FILEID");
            my $file   = $files_hashref->{$fileid};
            $file =~ m/^(\w+)/;
            my $prot = $1 if $1;
            $file =~ m/\/([^\/]+)$/; # find everything after the last slash
            my $filename = $1 if $1;

            # if the file is not a web reference, add 'file://' and the unpack path
            # to the filename
            for ($prot)
            {
              /http/ && do { last; };    # $filename stays unchanged

              { # default option
                $file =~ s#^#file://$unpack_dir/#;
                last;
              };
            } ## end for ($prot)

            $doc_data->{_parent} = $eprint
                unless defined $doc_data->{_parent};
            $doc_data->{eprintid} = $eprint->get_id
                unless defined $doc_data->{eprintid};
            $doc_data->{main} = $filename unless $doc_data->{main};
            $doc_data->{format} =
                $session->get_repository->call('guess_doc_type', $session,
                                               $filename)
                unless exists $doc_data->{format};

            my %file_data;
            $file_data{filename} = $filename;
            $file_data{url}      = $file;

            $doc_data->{files} = [] unless exists $doc_data->{files};
            push @{$doc_data->{files}}, \%file_data;
          } ## end foreach my $fptr (@nodes)
          
          my $doc_dataset = $session->get_repository->get_dataset("document");
          local $session->get_repository->{config}->{enable_web_imports}  = 1;
          local $session->get_repository->{config}->{enable_file_imports} = 1;
          $document = EPrints::DataObj::Document->create_from_data($session,
                                                     $doc_data, $doc_dataset);
     } ## end if (scalar @nodes)

  } ## end foreach my $element (@{$nodes_ref...})

  return $document;

} ## end sub process_struct_elements

sub add_file
{
  my ($eprint, $hash_ref, $unpack_dir) = @_;

  my $prot;

  my $file = $hash_ref->{'file'};
  $file =~ m/^(\w+)/;
  $prot = $1 if $1;
  $file =~ m/\/([^\/]+)$/; # find everything after the last slash
  my $filename = $1 if $1;

  return undef unless $filename;
  $filename =~ s# #\ #g; # replace spaces with '\ '
  # if the file is not a web reference, add 'file://' and the unpack path to the
  # filename
  for ($prot)
  {
    /http/ && do
    {

      # $filename stays unchanged
      last;
    };
    {
      $file =~ s#^#file://$unpack_dir/# if $file;
      last;
    };
  } ## end for ($prot)

  
  my $session = $hash_ref->{'session'};

  # Setup: we return a doc_data hash.
  # If one is passed in, we update & return that, else create a new one.
  my $doc_data = {};

  $doc_data->{_parent} = $eprint unless defined $doc_data->{_parent};
  $doc_data->{eprintid} = $eprint->get_id
      unless defined $doc_data->{eprintid};
  $doc_data->{main} = $filename unless $doc_data->{main};
  $doc_data->{format}
      = $session->get_repository->call('guess_doc_type', $session, $filename)
      unless exists $doc_data->{format};

  my %file_data;
  $file_data{filename} = $filename;
  $file_data{url}      = "$file";

  $doc_data->{files} = [\%file_data];

  return $doc_data;

} ## end sub add_file

sub keep_deposited_file
{
  return 1;
}

sub parse_epdcx_xml_data
{
  my ($plugin, $xml) = @_;

  my $set = ($xml->getElementsByTagName("descriptionSet"))[0];

  unless (defined $set)
  {
    $plugin->set_status_code(400);
    $plugin->add_verbose("ERROR: no <descriptionSet> tag found.");
    return;
  } ## end unless (defined $set)

  my $epdata = {};

  foreach my $desc ($set->getElementsByTagName("description"))
  {

    foreach my $stat ($desc->getElementsByTagName("statement"))
    {
      my ($field, $value) = _parse_statement($stat);

      if (defined $field)
      {
        if ($field eq 'date')
        {
          $epdata->{ispublished} = 'pub';
          $epdata->{date}        = $value;
          $epdata->{date_type}   = 'published';

        } elsif ($field eq 'creator')
        {

          $value = clean_text($value);

          if ($value =~ /(\w+)\,\s?(.*)$/)
          {
            push @{$epdata->{creators_name}}, {family => $1, given => $2};
          } else
          {
            push @{$epdata->{creators_name}}, {family => $value};
          }
        } ## end elsif ($field eq 'creator'...)
        elsif ($field eq 'cite')    # 'cite' defined in _parse_statement
        {

          my $cite = clean_text($value);

          # now we need to decode the citation
          # $publication [$vol[($num)]][, $pps|$pgs]
          my ($pub, $vol, $num, $pps, $pgs);

          # Start by pulling the page range or number of pages off the end
          $cite =~ s/\, (\d+(:?\-\d+))$//;
          $num = $1 if $1;
          if   ($num =~ /\-/) { $pps = $num }
          else                { $pgs = $num }
          $num = undef;

          # Do we have volume number ?
          $cite =~ s/\((.+)\)$//;
          $num = $1 if $1;

          # Get the volume number
          $cite =~ s/\s+(\d+)$//;
          $vol = $1 if $1;

          $pub = $cite;
          $epdata->{publication} = $pub if $pub;
          $epdata->{volume}      = $vol if $vol;
          $epdata->{number}      = $num if $num;
          $epdata->{pagerange}   = $pps if $pps;
          $epdata->{pages}       = $pgs if $pgs;
        } else
        {
          $epdata->{$field} = $value;
        }

      } ## end if (defined $field)

    } ## end foreach my $stat ($desc->getElementsByTagName...)
  } ## end foreach my $desc ($set->getElementsByTagName...)

  return $epdata;
} ## end sub parse_epdcx_xml_data

sub _parse_statement
{
  my ($stat) = @_;

  my $property = $stat->getAttribute("propertyURI");

  unless (defined $property)
  {
    $property = $stat->getAttribute("epdcx:propertyURI");
  }

  for ($property)
  {
    m#http://purl.org/dc/elements/1.1/type# && do
    {
      return _get_type($stat);
      last;
    };    ## end if ($property eq 'http://purl.org/dc/elements/1.1/type'...)

    m#http://purl.org/dc/elements/1.1/title# && do
    {
      my $value = ($stat->getElementsByTagName("valueString"))[0];
      my $title = EPrints::XML::to_string(EPrints::XML::contents_of($value));
      return ("title", clean_text($title));
      last;
    };  ## end elsif ($property eq 'http://purl.org/dc/elements/1.1/title'...)

    m#http://purl.org/dc/terms/abstract# && do
    {
      my $value = ($stat->getElementsByTagName("valueString"))[0];
      my $abstract
          = EPrints::XML::to_string(EPrints::XML::contents_of($value));
      return ("abstract", clean_text($abstract));
      last;
    };    ## end elsif ($property eq 'http://purl.org/dc/terms/abstract'...)

    m#http://purl.org/dc/elements/1.1/creator# && do
    {
      my $value = ($stat->getElementsByTagName("valueString"))[0];
      my $name  = EPrints::XML::to_string(EPrints::XML::contents_of($value));
      return ("creator", clean_text($name));
      last;
    }; ## end elsif ($property eq 'http://purl.org/dc/elements/1.1/creator'...)

    m#http://purl.org/dc/elements/1.1/identifier# && do
    {
      my $value = ($stat->getElementsByTagName("valueString"))[0];
      my $id    = EPrints::XML::to_string(EPrints::XML::contents_of($value));
      return ("id_number", clean_text($id));
      last;
    }; ## end elsif ($property eq 'http://purl.org/dc/elements/1.1/identifier'...)

    m#http://purl.org/eprint/terms/status# && do
    {
      if ($stat->getAttribute("valueURI") =~ /status\/(.*)$/)
      {
        my $status = $1;
        if ($status eq 'PeerReviewed')
        {
          $status = 'TRUE';
        } elsif ($status eq 'NonPeerReviewed')
        {
          $status = 'FALSE';
        } else
        {
          return;
        }

        return ('refereed', $status);    # is this the proper field?

      } ## end if ($stat->getAttribute...)

      return;
      last;
    };    ## end elsif ($property eq 'http://purl.org/eprint/terms/status'...)

    m#http://purl.org/dc/elements/1.1/language# && do
    {

      # LANGUAGE (not parsed)
      last;
    };

    m#http://purl.org/dc/terms/available# && do
    {
      my $value = ($stat->getElementsByTagName("valueString"))[0];
      my $id    = EPrints::XML::to_string(EPrints::XML::contents_of($value));
      return ("date", clean_text($id));
      last;
    };

    m#http://purl.org/eprint/terms/copyrightHolder# && do
    {

      # COPYRIGHT HOLDER (not parsed)
      last;
    };

    m#http://purl.org/dc/terms/bibliographicCitation# && do
    {
      my $value = ($stat->getElementsByTagName("valueString"))[0];
      my $cite  = EPrints::XML::to_string(EPrints::XML::contents_of($value));
      return ('cite', $cite);
      last;
    };

    m#http://purl.org/dc/elements/1.1/publisher# && do
    {
      my $value = ($stat->getElementsByTagName("valueString"))[0];
      $value = EPrints::XML::to_string(EPrints::XML::contents_of($value));
      return ('publisher', $value);
      last;
    };

    # This lot are OA-RJ extensions to cover blantently missing fields
    m#http://opendepot.org/broker/elements/1.0/publication# && do
    {
      my $value = ($stat->getElementsByTagName("valueString"))[0];
      $value = EPrints::XML::to_string(EPrints::XML::contents_of($value));
      return ('publication', $value);
      last;
    };

    m#http://opendepot.org/broker/elements/1.0/issn# && do
    {
      my $value = ($stat->getElementsByTagName("valueString"))[0];
      $value = EPrints::XML::to_string(EPrints::XML::contents_of($value));
      return ('issn', $value);
      last;
    };

    m#http://opendepot.org/broker/elements/1.0/isbn# && do
    {
      my $value = ($stat->getElementsByTagName("valueString"))[0];
      $value = EPrints::XML::to_string(EPrints::XML::contents_of($value));
      return ('isbn', $value);
      last;
    };

    m#http://opendepot.org/broker/elements/1.0/volume# && do
    {
      my $value = ($stat->getElementsByTagName("valueString"))[0];
      $value = EPrints::XML::to_string(EPrints::XML::contents_of($value));
      return ('volume', $value);
      last;
    };

    m#http://opendepot.org/broker/elements/1.0/issue# && do
    {
      my $value = ($stat->getElementsByTagName("valueString"))[0];
      $value = EPrints::XML::to_string(EPrints::XML::contents_of($value));
      return ('issue', $value);
      last;
    };

    m#http://opendepot.org/broker/elements/1.0/pages# && do
    {
      my $value = ($stat->getElementsByTagName("valueString"))[0];
      $value = EPrints::XML::to_string(EPrints::XML::contents_of($value));
      return ('pagerange', $value) if $value =~ /\-/;
      return ('pages', $value);
      last;
    };
  } ## end for ($property)
  return;

} ## end sub _parse_statement

sub _get_type
{
  my ($stat) = @_;

  if ($stat->getAttribute("valueURI") =~ /type\/(.*)$/)
  {    # then $1 = type

    # reference for these mappings is:
    # "http://www.ukoln.ac.uk/repositories/digirep/index/Eprints_Type_Vocabulary_Encoding_Scheme"

    my $type = $1;    # no need to clean_text( $1 ) here

    if (   $type eq 'JournalArticle'
        || $type eq 'JournalItem'
        || $type eq 'SubmittedJournalArticle'
        || $type eq 'WorkingPaper')
    {
      $type = 'article';
    } ## end if ($type eq 'JournalArticle'...)
    elsif ($type eq 'Book')
    {
      $type = 'book';
    } elsif ($type eq 'BookItem')
    {
      $type = 'book_section';
    } elsif (   $type eq 'ConferenceItem'
             || $type eq 'ConferencePoster'
             || $type eq 'ConferencePaper')
    {
      $type = 'conference_item';
    } ## end elsif ($type eq 'ConferenceItem'...)
    elsif ($type eq 'Patent')
    {
      $type = 'patent';
    } elsif ($type eq 'Report')
    {
      $type = 'monograph';    # I think?
    } elsif ($type eq 'Thesis')
    {
      $type = 'thesis';
    } else
    {
      return
          ; # problem there! But the user can still correct this piece of data...
    }

    return ("type", $type);
  } ## end if ($stat->getAttribute...)
  return;
} ## end sub _get_type

sub clean_text
{
  my ($text) = @_;

  my @lines = split("\n", $text);

  foreach (@lines)
  {
    $_ =~ s/\s+$//;

    $_ =~ s/^\s+//;
  } ## end foreach (@lines)

  return join(" ", @lines);
} ## end sub clean_text

1;

